﻿#ifndef ARBREGENERAL_HPP
#define ARBREGENERAL_HPP

#include <iostream>
#include <vector>
using namespace std;

template <class T> class ArbreGen {

private:     

  struct Node {
    T info;
    vector<Node*> seg;
  };

  Node* primer;

/*
Un árbol general puede ser un árbol vacío o un árbol no vacío que no
contiene ningún subárbol vacío.

Cada subárbol de un árbol general puede tener un número indeterminado
de hijos no vacíos, incluso cero hijos, pero ningún hijo vacío. 

Si un subárbol no tiene ningún hijo, entonces contiene un único nodo,
que es una hoja del subárbol y de los árboles que contienen a dicho
subárbol.

Los árboles generales se caracterizan por tener un número
indeterminado de hijos no vacíos que puede variar en cada subárbol.
*/


static Node* copia_jerarquia(Node* m) { 
/* Pre: cert */
/* Post: El resultat és NULL si m és NULL; en cas contrari, el
   resultat apunta al node arrel d'una jerarquia de nodes que és una
   còpia de la jerarquia de nodes que té el node apuntat per m com a
   arrel. */
Node* n;
if (m == NULL) n = NULL;
else {
  n = new Node;
  n->info = m->info;
  int ari = m->seg.size();
  n->seg = vector<Node*>(ari);
  for (int i=0; i<ari; ++i) { 
    n->seg[i] = copia_jerarquia(m->seg[i]); 
  }
}
return n;
}


static void esborra_jerarquia(Node* m) {  
/* Pre: Cert */
/* Post No fa res si m és NULL; en cas contrari, allibera espai de
   tots els nodes de la jerarquia que té el node apuntat per m com a
   arrel. */
if (m != NULL) {
  int ari = m->seg.size();
  for (int i=0; i<ari; ++i) 
    esborra_jerarquia(m->seg[i]);
  delete m;
}
}



public:


ArbreGen() {
/* Pre: cert */
/* Post: El p.i. és un arbre general buit. */
primer= NULL;
}


ArbreGen(const ArbreGen& original) {
/* Pre: cert */
/* Post: el p.i. és una còpia d'original */
primer = copia_jerarquia(original.primer);
}


ArbreGen(const T &x) {
/* Pre: cert */
/* Post: El p.i. és un arbre general amb arrel x i sense fills */
primer= new Node;
primer->info = x;
}


~ArbreGen()
// Destructora: Esborra automàticament els objectes locals en
// sortir d'un àmbit de visibilitat
{
esborra_jerarquia(primer);
}


ArbreGen& operator=(const ArbreGen& original)
/* Pre: cert */
/* Post: El p.i. passa a ser una còpia d'original i qualsevol
 contingut anterior del p.i. ha estat esborrat (excepte si el p.i. i
 original ja eren el mateix objecte). */
{
if (this != &original) {
  esborra_jerarquia(primer);
  primer = copia_jerarquia(original.primer);
}
return *this;
}


void a_buit() {
/* Pre: cert */
/* Post: el p.i. és un arbre general buit i qualsevol
   contingut anterior del p.i. ha estat esborrat */
esborra_jerarquia(primer);
primer = NULL;
}        


void plantar(const T &x) {
/* Pre: el p.i. és buit */
/* Post: el p.i. té x com a arrel i zero fills */
  primer= new Node;
  primer->info = x;
}

void plantar(const T &x, vector<ArbreGen> &v) {
/* Pre:  el p.i. és buit i v.size()>0 i cap arbre de v és buit */
/* Post: el p.i. té x com a arrel i els elements originals  
   de v com a fills, v passa a contenir arbres buits */
  primer = new Node;
  primer->info = x;
  int ari = v.size();
  primer->seg = vector<Node*>(ari);
  for (int i = 0; i < ari; ++i) {
    primer->seg[i] = v[i].primer;
    v[i].primer = NULL;
  }
}


void afegir_fill(const ArbreGen &a) {
/* Pre: el p.i. i a no són buits */
/* Post: el p.i. té un fill més que a l'inici, i aquest nou darrer fill 
   és una còpia de l'arbre a */
(primer->seg).push_back(copia_jerarquia(a.primer));
}       


void fill(const ArbreGen &a, int i) {
/* Pre: el p.i. és buit, a no és buit, i està entre 1 i el nombre de
   fills d'a */
/* Post: el p.i és una còpia del fill i-éssim d'a */
primer = copia_jerarquia((a.primer)->seg[i-1]);
}       


void fills(vector<ArbreGen> &v) {
/* Pre:  el p.i. no és buit i li diem A, v és un vector buit  */
/* Post: el p.i. és buit, v passa a contenir els fills de l'arbre A  */
  int ari = primer->seg.size();
  v = vector<ArbreGen>(ari);
  for (int i = 0; i < ari; ++i) {
     v[i].primer = primer->seg[i];
  }  
  delete primer;
  primer = NULL;
}


T arrel() const {
/* Pre: el p.i. no és buit */
/* Post: el resultat és l'arrel del p.i. */
 return primer->info;
}


bool es_buit() const {
/* Pre: cert */
/* Post: el resultat indica si el p.i. és un arbre buit */
return primer == NULL;
}


int nombre_fills() const {
/* Pre: el p.i. no és buit */
/* Post: el resultat és el nombre de fills del p.i. */
  return (primer->seg).size();
}



// --- Extensions -------------------------------------------------

#include "sustitucionG.hh" // ejemplo delete

#include "max_arbG.hh"

#include "sumaG_k.hh"

#include "buscaG.hh"

#include "arb_max_min.hh"  // ejemplo new
  
};

#endif
