﻿#ifndef LLISTA_HH
#define LLISTA_HH

#include <iostream> 
using namespace std;


template <class T> class Llista {

private:

struct Node2 {
  T info;
  Node2* seg;
  Node2* ant;
}; 

int longitud;
Node2* primer;
Node2* ultim;
Node2* act; // punto de interés

// No implementamos iteradores
// - sólo un punto de interés por lista
// - una Llista no se puede pasar por referencia constante  
//   en operaciones de búsqueda o recorrido.


static Node2* copia_cadena(Node2* m, Node2* oact,
				    Node2*& u, Node2*& a) { 
/* Pre: cert */
/* Post: si m és NULL, el resultat, u i a són NULL; en cas contrari,
 el resultat apunta al primer node d'una cadena de nodes que és 
 còpia de la cadena que té el node apuntat per m com a primer, u
 apunta a l'últim node, a és o bé NULL si oact no apunta a cap
 node de la cadena que comenca amb m o bé apunta al node còpia del
 node apuntat per oact */
Node2* n;
if (m==NULL) {n=NULL; u=NULL; a=NULL;}
else {
  n = new Node2;
  n->info = m->info;
  n->ant = NULL;
  n->seg = copia_cadena(m->seg, oact, u, a);
  if (n->seg == NULL) u = n; 
  else (n->seg)->ant = n;
  if (m == oact) a = n;
}
return n;
}



static void esborra_cadena(Node2* m) {     
/* Pre: cert */
/* Post: No fa res si m és NULL; en cas contrari, allibera espai dels
 nodes de la cadena que té el node apuntat per m com a primer. */
if (m != NULL) {
  esborra_cadena(m->seg);
  delete m;
}
}


public:


Llista() {
/* Pre: cert */
/* Post: el resultat es una llista buida */
longitud = 0;
primer = NULL;
ultim = NULL;
act = NULL;
}


Llista(const Llista &original) {
/* Pre: cert */
/* Post: El resultat és una còpia d'original */
longitud = original.longitud;
primer = copia_cadena(original.primer, original.act,
				ultim, act);  
}



~Llista() {
// Destructora: Esborra automaticament els objectes locals en
// sortir d'un ambit de visibilitat
esborra_cadena(primer);
}


Llista& operator=(const Llista& original) { 
/* Pre: cert */
/* Post: El p.i. passa a ser una copia d'original i qualsevol contingut
 anterior del p.i. ha estat esborrat (excepte si el p.i. i original ja
 eren el mateix objecte) */
if (this != &original) {
  longitud = original.longitud;
  esborra_cadena(primer);
  primer = copia_cadena(original.primer, original.act,
				  ultim, act);
}
return *this;
}


// ¿Cómo se llama en la clase list de STL?
void l_buida() {
/* Pre: cert */
/* Post: El p.i. passa a ser una llista sense elements i qualsevol
 contingut anterior del p.i. ha estat esborrat */
esborra_cadena(primer);
longitud = 0;
primer = NULL;
ultim = NULL;
act = NULL;
}



void afegir(const T& x) {
/* Pre: cert */
/* Post: el p.i. és com el seu valor original, però amb x afegit a
 l'esquerra del punt d'interès */
Node2* aux = new Node2; 
aux->info = x;
aux->seg = act;
 
if (longitud==0) {
  aux->ant = NULL;
  primer = aux;
  ultim = aux;
}
else if (act==NULL) {
  aux->ant = ultim;
  ultim->seg = aux;
  ultim = aux;
}
else if (act==primer) {
  aux->ant = NULL;
  primer->ant = aux; 
  primer = aux;
}
else {
  aux->ant = act->ant;
  (act->ant)->seg = aux;
  act->ant = aux;
}
 
++longitud;
}


void eliminar() {
/* Pre: el p.i. és una llista no buida i el seu punt d'interès no està a
 la dreta de tot */
/* Post: El p.i. és com el p.i. original sense l'element on estava el
 punt d'interès i amb el nou punt d'interès avançat una posició a la
 dreta */
Node2* aux = act; 

if (longitud==1) {
  primer = NULL;
  ultim = NULL;
}
else if (act==primer) {
  primer = primer->seg;
  primer->ant = NULL;
}
else if (act==ultim) {
  ultim = ultim->ant;
  ultim->seg = NULL;
}
else {
  (act->ant)->seg = act->seg;
  (act->seg)->ant = act->ant;
}

act = act->seg; 
delete aux; 
--longitud;
}


void concat(Llista& l) {
/* Pre: l = L; l no és el mateix objecte que el p.i. */
/* Post: el p.i. té els seus elements originals seguits pels de L, l és
 buida i el punt d'interès del p.i. queda situat a l'inici */
if (this != &l) { // Les dues llistes han de ser objectes diferents.
  if (l.longitud > 0) { // Si la llista l és buida no cal fer res.
    if (longitud == 0) {
      primer = l.primer;
      ultim = l.ultim;
      longitud = l.longitud;
    }
    else {
      ultim->seg = l.primer;
      (l.primer)->ant = ultim;
      ultim = l.ultim;
      longitud += l.longitud;
    }
    l.primer = NULL;
    l.ultim = NULL;
    l.act = NULL;
    l.longitud = 0;  
  }    
  act = primer;
}
}


bool es_buida() const {
/* Pre: cert */
/* Post: El resultat indica si el p.i. és una llista buida o no */
return longitud == 0;
}


int mida() const {
/* Pre: cert */
/* Post: El resultat és el nombre d'elements de la llista p.i. */
return longitud;
}


T actual() const { 
/* Pre: el p.i. és una llista no buida i el seu punt d'interès no
 està a la dreta de tot */
/* Post: el resultat és l'element actual del p.i. */
return act->info;
}


void modifica_actual(const T &x) { 
/* Pre: el p.i. és una llista no buida i el seu punt d'interès no
 està a la dreta de tot */
/* Post: el p.i. és com el seu valor original, pero amb x
 reemplaçant l'element actual */
act->info = x; 
}


void inici() {
/* Pre: cert */
/* Post: el punt d'interès del p.i. està situat a sobre del primer
 element de la llista o a la dreta de tot si la llista és buida */
act = primer; 
}


void fi() {
/* Pre: cert */
/* Post: el punt d'interes del p.i. esta situat a la dreta del tot */
act = NULL; 
}


void avanca() { 
/* Pre: el punt d'interes del p.i. no estó a la dreta de tot */
/* Post: el punt d'interes del p.i. esta situat una posicio mes a la
 dreta que al valor original del p.i. */ 
act = act->seg; 
}


void retrocedeix() {
/* Pre: el punt d'interes del p.i. no està a sobre del primer
 element de la llista */
/* Post: el punt d'interes del p.i. esta situat una posicio mes a
 l'esquerra que al valor original del p.i. */ 
if (act == NULL) act = ultim;  
else act = act->ant; 
}


bool dreta_de_tot() const {
/* Pre: cert */
/* Post: el resultat indica si el punt d'interes del p.i. esta a la
 dreta de tot */
return act == NULL; 
}


bool sobre_el_primer() const {
/* Pre: cert */
/* Post: el resultat indica si el punt d'interós del p.i. esta a sobre del
primer element del p.i. o esta a la dreta de tot si la llista es buida */
return act == primer; 
}


// Extensio ------------------------------------------------

//------- Invertir el orden de los elementos de una lista ----- 
// Usamos la idea de intercambiar los punteros anterior y siguiente en
// cada nodo.

void invertir() {
/* Pre: cert */ 
/* Post: el p.i. té els mateixos elements que a l'inici però amb
 l'ordre invertit i el seu punt d'interès no s'ha mogut */
// c = C es la cadena de nodos enlazados a la que apunta n. 
Node2* n = primer;
/* Inv: c es una cadena de nodos enlazados con los mismos elementos
que C, excepto que se han intercambiado los punteros ant y seg de los
nodos anteriores en C al nodo al que apunta n. */    
while (n != ultim) {
  Node2* aux = n->seg;
  n->seg = n->ant;
  n->ant = aux;

  n = aux; // avanza n, aux es el seg de n en C
}

if (n != NULL and n != primer) {
  n->seg = n->ant; // Tratamos el último nodo de C si existe y 
  n->ant = NULL;   // es distinto del primero de C.
  ultim = primer; // El último de C es el primero de c.
  primer = n;     // El primero de c es el último de C.
}
}


};
#endif
